/*
 * beetest.c
 *
 * Beecrypt test and benchmark application
 *
 * Copyright (c) 1999-2000, Virtual Unlimited B.V.
 *
 * Author: Bob Deblier <bob@virtualunlimited.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "beecrypt.h"
#include "blockmode.h"
#include "mp32barrett.h"
#include "dldp.h"

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <stdio.h>

void testBlockInit(uint8* block, int length)
{
	register int i;
	for (i = 1; i <= length; i++)
		*(block++) = (uint8) i;
}

void testBlockCiphers()
{
	int i, j, k;

	printf("\tTesting the blockciphers:\n");

	for (i = 0; i < blockCipherCount(); i++)
	{
		const blockCipher* tmp = blockCipherGet(i);

		if (tmp)
		{
			uint32 blockwords = tmp->blockbits >> 5;

			uint32* src_block = (uint32*) malloc(blockwords * sizeof(uint32));
			uint32* dst_block = (uint32*) malloc(blockwords * sizeof(uint32));
			uint32* spd_block = (uint32*) malloc(1024 * 1024 * blockwords * sizeof(uint32));

			void* encrypt_param = (void*) malloc(tmp->paramsize);
			void* decrypt_param = (void*) malloc(tmp->paramsize);

			printf("\t%s:\n", tmp->name);

			for (k = tmp->keybitsmin; k <= tmp->keybitsmax; k += tmp->keybitsinc)
			{
				void* key = (void*) malloc(k >> 3);

				testBlockInit((uint8*) key, k >> 3);

				printf("\t\tsetup encrypt (%d bits key): ", k);
				if (tmp->setup(encrypt_param, key, k, ENCRYPT) < 0)
				{
					free(key);
					printf("failed\n");
					continue;
				}
				printf("ok\n");
				printf("\t\tsetup decrypt (%d bits key): ", k);
				if (tmp->setup(decrypt_param, key, k, DECRYPT) < 0)
				{
					free(key);
					printf("failed\n");
					continue;
				}
				printf("ok\n");
				printf("\t\tencrypt/decrypt test block: ");
				testBlockInit((uint8*) src_block, tmp->blockbits >> 3);
				memcpy(dst_block, src_block, tmp->blockbits >> 3);
				tmp->encrypt(encrypt_param, dst_block);
				/*
				for (j = 0; j < (tmp->blockbits >> 3); j++)
				{
					printf("%02x", *(((uint8*)dst_block)+j));
				}
				printf(" ");
				*/
				tmp->decrypt(decrypt_param, dst_block);
				if (memcmp(src_block, dst_block, tmp->blockbits >> 3))
				{
					free(key);
					printf("failed\n");
					continue;
				}
				free(key);
				printf("ok\n");
				printf("\t\tspeed measurement:\n");
				{
					#if HAVE_SYS_TIME_H
					long sec, usec;
					struct timeval tstart, tstop;
					#endif

					#if HAVE_SYS_TIME_H
					gettimeofday(&tstart, 0);
					#endif
					blockEncrypt(tmp, encrypt_param, ECB, 1024 * 1024, spd_block, spd_block, 0);
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstop, 0);
					sec = tstop.tv_sec - tstart.tv_sec;
					usec = tstop.tv_usec - tstart.tv_usec;
					if (usec < 0)
					{
						sec--;
						usec += 1000000;
					}
					printf("\t\t\tECB encrypts 1M blocks of %d bits in %d.%03d seconds\n", tmp->blockbits, sec, usec / 1000);
					#endif
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstart, 0);
					#endif
					blockDecrypt(tmp, decrypt_param, ECB, 1024 * 1024, spd_block, spd_block, 0);
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstop, 0);
					sec = tstop.tv_sec - tstart.tv_sec;
					usec = tstop.tv_usec - tstart.tv_usec;
					if (usec < 0)
					{
						sec--;
						usec += 1000000;
					}
					printf("\t\t\tECB decrypts 1M blocks of %d bits in %d.%03d seconds\n", tmp->blockbits, sec, usec / 1000);
					#endif
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstart, 0);
					#endif
					blockEncrypt(tmp, encrypt_param, CBC, 1024 * 1024, spd_block, spd_block, 0);
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstop, 0);
					sec = tstop.tv_sec - tstart.tv_sec;
					usec = tstop.tv_usec - tstart.tv_usec;
					if (usec < 0)
					{
						sec--;
						usec += 1000000;
					}
					printf("\t\t\tCBC encrypts 1M blocks of %d bits in %d.%03d seconds\n", tmp->blockbits, sec, usec / 1000);
					#endif
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstart, 0);
					#endif
					blockEncrypt(tmp, decrypt_param, CBC, 1024 * 1024, spd_block, spd_block, 0);
					#if HAVE_SYS_TIME_H
					gettimeofday(&tstop, 0);
					sec = tstop.tv_sec - tstart.tv_sec;
					usec = tstop.tv_usec - tstart.tv_usec;
					if (usec < 0)
					{
						sec--;
						usec += 1000000;
					}
					printf("\t\t\tCBC decrypts 1M blocks of %d bits in %d.%03d seconds\n", tmp->blockbits, sec, usec / 1000);
					#endif
				}
			}
			free(spd_block);
			free(dst_block);
			free(src_block);
			free(decrypt_param);
			free(encrypt_param);
		}
	}
}

void testHashFunctions()
{
	int i, j, k;

	uint8* data = (uint8*) malloc(16 * 1024 * 1024);

	if (data)
	{
		hashFunctionContext hfc;

		printf("\tTesting the hash functions:\n");

		for (i = 0; i < hashFunctionCount(); i++)
		{
			const hashFunction* tmp = hashFunctionGet(i);

			if (tmp)
			{
				uint8* digest = (uint8*) malloc(tmp->digestsize);

				printf("\t%s:\n", tmp->name);

				if (digest)
				{
					#if HAVE_SYS_TIME_H
					long sec, usec;
					struct timeval tstart, tstop;
					#endif

					hashFunctionContextInit(&hfc, tmp);

					#if HAVE_SYS_TIME_H
					gettimeofday(&tstart, 0);
					#endif
					hfc.hash->reset(hfc.param);
					hfc.hash->update(hfc.param, data, 16 * 1024 * 1024);
					hfc.hash->digest(hfc.param, (uint32*) digest);

					#if HAVE_SYS_TIME_H
					gettimeofday(&tstop, 0);
					sec = tstop.tv_sec - tstart.tv_sec;
					usec = tstop.tv_usec - tstart.tv_usec;
					if (usec < 0)
					{
						sec--;
						usec += 1000000;
					}
					printf("\t\thashes 16MB in %d.%03d seconds\n", sec, usec / 1000);
					#endif

					free(digest);
				}
				hashFunctionContextFree(&hfc);
			}
		}
	}
}

void testExpMods()
{
	static const char* x_160 = "a361fc0d3d4513557cb8404a053313a51ee2278b";
	static const char* p_512 = "ffcf0a0767f18f9b659d92b9550351430737c3633dc6ae7d52445d937d8336e07a7ccdb119e9ab3e011a8f938151230e91187f84ac05c3220f335193fc5e351b";

	static const char* p_768 = "f9c3dc0b8e199094e3e69386e01de863908348196d6ad2557065e6ba36d10412579f394d1114c954ee647c84551d52f214e1e1682a75e7074b91085cfaf20b2888aa056bf760948a0b678bc253633eccfca86556ddb90f000ef93041b0d53171";

	static const char* p_1024 = "c615c47a56b47d869010256171ab164525f2ef4b887a4e0cdfc87043a9dd8894f2a18fa56729448e700f4b7420470b61257d11ecefa9ff518dc9fed5537ec6a9665ba73c948674320ff61b29c4cfa61e5baf47dfc1b80939e1bffb51787cc3252c4d1190a7f13d1b0f8d4aa986571ce5d4de5ecede1405e9bc0b5bf040a46d99";

	randomGeneratorContext rc;

	mp32barrett p;
	mp32number tmp;
	mp32number g;
	mp32number x;

	mp32bzero(&p);
	mp32nzero(&g);
	mp32nzero(&x);
	mp32nzero(&tmp);

	mp32nsethex(&x, x_160);

	randomGeneratorContextInit(&rc, randomGeneratorDefault());

	if (rc.rng && rc.param)
	{
		if (rc.rng->setup(rc.param) == 0)
		{
			int i;
			#if HAVE_SYS_TIME_H
			long sec, usec;
			struct timeval tstart, tstop;
			#endif
			printf("Timing modular exponentiations\n");
			printf("\t(512 bits ^ 160 bits) mod 512 bits:");
			mp32nsethex(&tmp, p_512);
			mp32bset(&p, tmp.size, tmp.data);
			mp32nsize(&g, p.size);
			mp32brndres(&p, g.data, &rc);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstart, 0);
			#endif
			for (i = 0; i < 100; i++)
				mp32bnpowmod(&p, &g, &x);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstop, 0);
			sec = tstop.tv_sec - tstart.tv_sec;
			usec = tstop.tv_usec - tstart.tv_usec;
			if (usec < 0)
			{
				sec--;
				usec += 1000000;
			}
			printf("\t 100x in %d.%03d seconds\n", sec, usec / 1000);
			#endif
			printf("\t(768 bits ^ 160 bits) mod 768 bits:");
			mp32nsethex(&tmp, p_768);
			mp32bset(&p, tmp.size, tmp.data);
			mp32nsize(&g, p.size);
			mp32brndres(&p, g.data, &rc);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstart, 0);
			#endif
			for (i = 0; i < 100; i++)
				mp32bnpowmod(&p, &g, &x);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstop, 0);
			sec = tstop.tv_sec - tstart.tv_sec;
			usec = tstop.tv_usec - tstart.tv_usec;
			if (usec < 0)
			{
				sec--;
				usec += 1000000;
			}
			printf("\t 100x in %d.%03d seconds\n", sec, usec / 1000);
			#endif
			printf("\t(1024 bits ^ 160 bits) mod 1024 bits:");
			mp32nsethex(&tmp, p_1024);
			mp32bset(&p, tmp.size, tmp.data);
			mp32nsize(&g, p.size);
			mp32brndres(&p, g.data, &rc);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstart, 0);
			#endif
			for (i = 0; i < 100; i++)
				mp32bnpowmod(&p, &g, &x);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstop, 0);
			sec = tstop.tv_sec - tstart.tv_sec;
			usec = tstop.tv_usec - tstart.tv_usec;
			if (usec < 0)
			{
				sec--;
				usec += 1000000;
			}
			printf("\t 100x in %d.%03d seconds\n", sec, usec / 1000);
			#endif
			mp32bfree(&p);
			mp32nfree(&g);
			mp32nfree(&x);
			mp32nfree(&tmp);
		}
	}

	randomGeneratorContextFree(&rc);
}

void testDLParams()
{
	randomGeneratorContext rc;
	dldp_p dp;

	memset(&dp, 0, sizeof(dldp_p));

	randomGeneratorContextInit(&rc, randomGeneratorDefault());

	if (rc.rng && rc.param)
	{
		if (rc.rng->setup(rc.param) == 0)
		{
			#if HAVE_SYS_TIME_H
			long sec, usec;
			struct timeval tstart, tstop;
			gettimeofday(&tstart, 0);
			#endif
			printf("Generating P (768 bits) Q (512 bits) G with order Q\n");
			dldp_pgoqMake(&dp, &rc, 768 >> 5, 512 >> 5, 1);
			#if HAVE_SYS_TIME_H
			gettimeofday(&tstop, 0);
			sec = tstop.tv_sec - tstart.tv_sec;
			usec = tstop.tv_usec - tstart.tv_usec;
			if (usec < 0)
			{
				sec--;
				usec += 1000000;
			}
			printf("\tdone in %d.%03d seconds\n", sec, usec / 1000);
			#endif
		}
	}

	randomGeneratorContextFree(&rc);
}

main()
{
	int i, j;
	printf("the beecrypt library implements:\n");
	printf("\t%d entropy source%s:\n", entropySourceCount(), entropySourceCount() == 1 ? "" : "s");
	for (i = 0; i < entropySourceCount(); i++)
	{
		const entropySource* tmp = entropySourceGet(i);
		if (tmp)
			printf("\t\t%s\n", tmp->name);
		else
			printf("*** error: library corrupt\n");
	}
	printf("\t%d random generator%s:\n", randomGeneratorCount(), randomGeneratorCount() == 1 ? "" : "s");
	for (i = 0; i < randomGeneratorCount(); i++)
	{
		const randomGenerator* tmp = randomGeneratorGet(i);
		if (tmp)
			printf("\t\t%s\n", tmp->name);
		else
			printf("*** error: library corrupt\n");
	}
	printf("\t%d hash function%s:\n", hashFunctionCount(), hashFunctionCount() == 1 ? "" : "s");
	for (i = 0; i < hashFunctionCount(); i++)
	{
		const hashFunction* tmp = hashFunctionGet(i);
		if (tmp)
			printf("\t\t%s\n", tmp->name);
		else
			printf("*** error: library corrupt\n");
	}
	printf("\t%d keyed hash function%s:\n", keyedHashFunctionCount(), keyedHashFunctionCount() == 1 ? "" : "s");
	for (i = 0; i < keyedHashFunctionCount(); i++)
	{
		const keyedHashFunction* tmp = keyedHashFunctionGet(i);
		if (tmp)
			printf("\t\t%s\n", tmp->name);
		else
			printf("*** error: library corrupt\n");
	}
	printf("\t%d blockcipher%s:\n", blockCipherCount(), blockCipherCount() == 1 ? "" : "s");
	for (i = 0; i < blockCipherCount(); i++)
	{
		const blockCipher* tmp = blockCipherGet(i);
		if (tmp)
		{
			printf("\t\t%s ", tmp->name);
			for (j = tmp->keybitsmin; j <= tmp->keybitsmax; j += tmp->keybitsinc)
			{
				printf("%d", j);
				if (j < tmp->keybitsmax)
					printf("/");
				else
					printf(" bit keys\n");
			}
		}
		else
			printf("*** error: library corrupt\n");
	}

	testBlockCiphers();
	testHashFunctions();
	testExpMods();
	testDLParams();
}
